# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""Signature Version 4 test suite.

AWS provides a test suite for signature version 4:

https://github.com/awslabs/aws-c-auth/tree/v0.3.15/tests/aws-sig-v4-test-suite

This module contains logic to run these tests.  The test files were
placed in ./aws4_testsuite, and we're using those to dynamically
generate testcases based on these files.

"""

import datetime
import io
import logging
import os
import re
from http.server import BaseHTTPRequestHandler

import pytest

import botocore.auth
from botocore.awsrequest import AWSRequest
from botocore.compat import parse_qsl, urlsplit
from botocore.credentials import Credentials
from tests import FreezeTime

SECRET_KEY = "wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY"
ACCESS_KEY = 'AKIDEXAMPLE'
DATE = datetime.datetime(2015, 8, 30, 12, 36, 0, tzinfo=datetime.timezone.utc)
SERVICE = 'service'
REGION = 'us-east-1'

TESTSUITE_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'aws4_testsuite'
)

# The following tests are not run.  Each test has a comment as
# to why the test is being ignored.
TESTS_TO_IGNORE = [
    # Bad request-line syntax, python's HTTP parser chokes on this.
    'normalize-path/get-space',
    # Multiple query params of the same key not supported by the SDKs.
    'get-vanilla-query-order-key-case',
    'get-vanilla-query-order-key',
    'get-vanilla-query-order-value',
]

log = logging.getLogger(__name__)


class RawHTTPRequest(BaseHTTPRequestHandler):
    def __init__(self, raw_request):
        if isinstance(raw_request, str):
            raw_request = raw_request.encode('utf-8')
        self.rfile = io.BytesIO(raw_request)
        self.raw_requestline = self.rfile.readline()
        self.error_code = None
        self.error_message = None
        self.parse_request()

    def send_error(self, code, message):
        self.error_code = code
        self.error_message = message


def generate_test_cases():
    for dirpath, dirnames, filenames in os.walk(TESTSUITE_DIR):
        if not any(f.endswith('.req') for f in filenames):
            continue

        test_case = os.path.relpath(dirpath, TESTSUITE_DIR).replace(
            os.sep, '/'
        )
        if test_case in TESTS_TO_IGNORE:
            log.debug("Skipping test: %s", test_case)
            continue

        yield test_case


@pytest.mark.parametrize("test_case", generate_test_cases())
@FreezeTime(module=botocore.auth.datetime, date=DATE)
def test_signature_version_4(test_case):
    _test_signature_version_4(test_case)


def create_request_from_raw_request(raw_request):
    request = AWSRequest()
    raw = RawHTTPRequest(raw_request)
    if raw.error_code is not None:
        raise Exception(raw.error_message)
    request.method = raw.command
    datetime_now = DATE
    request.context['timestamp'] = datetime_now.strftime('%Y%m%dT%H%M%SZ')
    for key, val in raw.headers.items():
        request.headers[key] = val
    request.data = raw.rfile.read()
    host = raw.headers.get('host', '')
    # For whatever reason, the BaseHTTPRequestHandler encodes
    # the first line of the response as 'iso-8859-1',
    # so we need decode this into utf-8.
    if isinstance(raw.path, str):
        raw.path = raw.path.encode('iso-8859-1').decode('utf-8')
    url = f'https://{host}{raw.path}'
    if '?' in url:
        split_url = urlsplit(url)
        params = dict(parse_qsl(split_url.query))
        request.url = split_url.path
        request.params = params
    else:
        request.url = url
    return request


def _test_signature_version_4(test_case):
    test_case = SignatureTestCase(test_case)
    request = create_request_from_raw_request(test_case.raw_request)

    auth = botocore.auth.SigV4Auth(test_case.credentials, SERVICE, REGION)
    actual_canonical_request = auth.canonical_request(request)
    actual_string_to_sign = auth.string_to_sign(
        request, actual_canonical_request
    )
    auth.add_auth(request)
    actual_auth_header = request.headers['Authorization']

    # Some stuff only works right when you go through auth.add_auth()
    # So don't assert the interim steps unless the end result was wrong.
    if actual_auth_header != test_case.authorization_header:
        assert_equal(
            actual_canonical_request,
            test_case.canonical_request,
            test_case.raw_request,
            'canonical_request',
        )

        assert_equal(
            actual_string_to_sign,
            test_case.string_to_sign,
            test_case.raw_request,
            'string_to_sign',
        )

        assert_equal(
            actual_auth_header,
            test_case.authorization_header,
            test_case.raw_request,
            'authheader',
        )


def assert_equal(actual, expected, raw_request, part):
    if actual != expected:
        message = f"The {part} did not match"
        message += f"\nACTUAL:{actual!r} !=\nEXPECT:{expected!r}"
        message += f'\nThe raw request was:\n{raw_request}'
        raise AssertionError(message)


class SignatureTestCase:
    def __init__(self, test_case):
        filepath = os.path.join(
            TESTSUITE_DIR, test_case, os.path.basename(test_case)
        )
        self.raw_request = open(filepath + '.req', encoding='utf-8').read()
        self.canonical_request = (
            open(filepath + '.creq', encoding='utf-8').read().replace('\r', '')
        )
        self.string_to_sign = (
            open(filepath + '.sts', encoding='utf-8').read().replace('\r', '')
        )
        self.authorization_header = (
            open(filepath + '.authz', encoding='utf-8')
            .read()
            .replace('\r', '')
        )
        self.signed_request = open(filepath + '.sreq', encoding='utf-8').read()

        token_pattern = r'^x-amz-security-token:(.*)$'
        token_match = re.search(
            token_pattern, self.canonical_request, re.MULTILINE
        )
        token = token_match.group(1) if token_match else None
        self.credentials = Credentials(ACCESS_KEY, SECRET_KEY, token)



#!/usr/bin/env
# Copyright (c) 2012-2013 Mitch Garnaat http://garnaat.org/
# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore import retryhandler
from botocore.exceptions import (
    ChecksumError,
    EndpointConnectionError,
    ReadTimeoutError,
)
from tests import mock, unittest

HTTP_500_RESPONSE = mock.Mock()
HTTP_500_RESPONSE.status_code = 500

HTTP_400_RESPONSE = mock.Mock()
HTTP_400_RESPONSE.status_code = 400

HTTP_200_RESPONSE = mock.Mock()
HTTP_200_RESPONSE.status_code = 200

REQUEST_DICT = {'context': {}}


class TestRetryCheckers(unittest.TestCase):
    def construct_checker_kwargs(
        self, response, attempt_number, caught_exception
    ):
        checker_kwargs = {
            'attempt_number': attempt_number,
            'response': response,
            'caught_exception': caught_exception,
        }
        if isinstance(self.checker, retryhandler.MaxAttemptsDecorator):
            checker_kwargs.update({'retries_context': REQUEST_DICT['context']})

        return checker_kwargs

    def assert_should_be_retried(
        self, response, attempt_number=1, caught_exception=None
    ):
        checker_kwargs = self.construct_checker_kwargs(
            response, attempt_number, caught_exception
        )

        self.assertTrue(self.checker(**checker_kwargs))

    def assert_should_not_be_retried(
        self, response, attempt_number=1, caught_exception=None
    ):
        checker_kwargs = self.construct_checker_kwargs(
            response, attempt_number, caught_exception
        )

        self.assertFalse(self.checker(**checker_kwargs))

    def test_status_code_checker(self):
        self.checker = retryhandler.HTTPStatusCodeChecker(500)
        self.assert_should_be_retried(response=(HTTP_500_RESPONSE, {}))

    def test_max_attempts(self):
        self.checker = retryhandler.MaxAttemptsDecorator(
            retryhandler.HTTPStatusCodeChecker(500), max_attempts=3
        )
        response = {'ResponseMetadata': {}}

        # Retry up to three times.
        self.assert_should_be_retried(
            (HTTP_500_RESPONSE, response), attempt_number=1
        )
        self.assert_should_be_retried(
            (HTTP_500_RESPONSE, {}), attempt_number=2
        )
        # On the third failed response, we've reached the
        # max attempts so we should return False.
        self.assert_should_not_be_retried(
            (HTTP_500_RESPONSE, response), attempt_number=3
        )
        self.assertTrue(response['ResponseMetadata']['MaxAttemptsReached'])

    def test_max_attempts_successful(self):
        self.checker = retryhandler.MaxAttemptsDecorator(
            retryhandler.HTTPStatusCodeChecker(500), max_attempts=3
        )

        self.assert_should_be_retried(
            (HTTP_500_RESPONSE, {}), attempt_number=1
        )
        # The second retry is successful.
        self.assert_should_not_be_retried(
            (HTTP_200_RESPONSE, {}), attempt_number=2
        )

        # But now we can reuse this object.
        self.assert_should_be_retried(
            (HTTP_500_RESPONSE, {}), attempt_number=1
        )
        self.assert_should_be_retried(
            (HTTP_500_RESPONSE, {}), attempt_number=2
        )
        self.assert_should_not_be_retried(
            (HTTP_500_RESPONSE, {}), attempt_number=3
        )

    def test_error_code_checker(self):
        self.checker = retryhandler.ServiceErrorCodeChecker(
            status_code=400, error_code='Throttled'
        )
        response = (HTTP_400_RESPONSE, {'Error': {'Code': 'Throttled'}})
        self.assert_should_be_retried(response)

    def test_error_code_checker_does_not_match(self):
        self.checker = retryhandler.ServiceErrorCodeChecker(
            status_code=400, error_code='Throttled'
        )
        response = (HTTP_400_RESPONSE, {'Error': {'Code': 'NotThrottled'}})
        self.assert_should_not_be_retried(response)

    def test_error_code_checker_ignore_caught_exception(self):
        self.checker = retryhandler.ServiceErrorCodeChecker(
            status_code=400, error_code='Throttled'
        )
        self.assert_should_not_be_retried(
            response=None, caught_exception=RuntimeError()
        )

    def test_multi_checker(self):
        checker = retryhandler.ServiceErrorCodeChecker(
            status_code=400, error_code='Throttled'
        )
        checker2 = retryhandler.HTTPStatusCodeChecker(500)
        self.checker = retryhandler.MultiChecker([checker, checker2])
        self.assert_should_be_retried((HTTP_500_RESPONSE, {}))
        self.assert_should_be_retried(
            response=(HTTP_400_RESPONSE, {'Error': {'Code': 'Throttled'}})
        )
        self.assert_should_not_be_retried(response=(HTTP_200_RESPONSE, {}))

    def test_exception_checker_ignores_response(self):
        self.checker = retryhandler.ExceptionRaiser()
        self.assert_should_not_be_retried(
            response=(HTTP_200_RESPONSE, {}), caught_exception=None
        )

    def test_value_error_raised_when_missing_response_and_exception(self):
        self.checker = retryhandler.ExceptionRaiser()
        with self.assertRaises(ValueError):
            self.checker(1, response=None, caught_exception=None)


class TestCreateRetryConfiguration(unittest.TestCase):
    def setUp(self):
        self.retry_config = {
            '__default__': {
                'max_attempts': 5,
                'delay': {
                    'type': 'exponential',
                    'base': 1,
                    'growth_factor': 2,
                },
                'policies': {
                    'throttling': {
                        'applies_when': {
                            'response': {
                                'service_error_code': 'Throttling',
                                'http_status_code': 400,
                            }
                        }
                    }
                },
            },
            'OperationFoo': {
                'policies': {
                    'crc32check': {
                        'applies_when': {
                            'response': {
                                'crc32body': 'x-amz-crc32',
                            }
                        }
                    }
                }
            },
            'OperationBar': {
                'policies': {
                    'socket_errors': {
                        'applies_when': {
                            'socket_errors': ["GENERAL_CONNECTION_ERROR"],
                        }
                    }
                }
            },
        }

    def test_create_retry_single_checker_service_level(self):
        checker = retryhandler.create_checker_from_retry_config(
            self.retry_config, operation_name=None
        )
        self.assertIsInstance(checker, retryhandler.MaxAttemptsDecorator)
        # We're reaching into internal fields here, but only to check
        # that the object is created properly.
        self.assertEqual(checker._max_attempts, 5)
        self.assertIsInstance(
            checker._checker, retryhandler.ServiceErrorCodeChecker
        )
        self.assertEqual(checker._checker._error_code, 'Throttling')
        self.assertEqual(checker._checker._status_code, 400)

    def test_create_retry_for_operation(self):
        checker = retryhandler.create_checker_from_retry_config(
            self.retry_config, operation_name='OperationFoo'
        )
        self.assertIsInstance(checker, retryhandler.MaxAttemptsDecorator)
        self.assertEqual(checker._max_attempts, 5)
        self.assertIsInstance(checker._checker, retryhandler.MultiChecker)

    def test_retry_with_socket_errors(self):
        checker = retryhandler.create_checker_from_retry_config(
            self.retry_config, operation_name='OperationBar'
        )
        self.assertIsInstance(checker, retryhandler.BaseChecker)
        all_checkers = checker._checker._checkers
        self.assertIsInstance(
            all_checkers[0], retryhandler.ServiceErrorCodeChecker
        )
        self.assertIsInstance(all_checkers[1], retryhandler.ExceptionRaiser)

    def test_create_retry_handler_with_socket_errors(self):
        handler = retryhandler.create_retry_handler(
            self.retry_config, operation_name='OperationBar'
        )
        exception = EndpointConnectionError(endpoint_url='')
        with self.assertRaises(EndpointConnectionError):
            handler(
                response=None,
                attempts=10,
                caught_exception=exception,
                request_dict=REQUEST_DICT,
            )
        # No connection error raised because attempts < max_attempts.
        sleep_time = handler(
            response=None,
            attempts=1,
            caught_exception=exception,
            request_dict=REQUEST_DICT,
        )
        self.assertEqual(sleep_time, 1)
        # But any other exception should be raised even if
        # attempts < max_attempts.
        with self.assertRaises(ValueError):
            sleep_time = handler(
                response=None,
                attempts=1,
                caught_exception=ValueError(),
                request_dict=REQUEST_DICT,
            )

    def test_connection_timeouts_are_retried(self):
        # If a connection times out, we get a Timout exception
        # from requests.  We should be retrying those.
        handler = retryhandler.create_retry_handler(
            self.retry_config, operation_name='OperationBar'
        )
        sleep_time = handler(
            response=None,
            attempts=1,
            caught_exception=ReadTimeoutError(endpoint_url=''),
            request_dict=REQUEST_DICT,
        )
        self.assertEqual(sleep_time, 1)

    def test_create_retry_handler_with_no_operation(self):
        handler = retryhandler.create_retry_handler(
            self.retry_config, operation_name=None
        )
        self.assertIsInstance(handler, retryhandler.RetryHandler)
        # No good way to test for the delay function as the action
        # other than to just invoke it.
        self.assertEqual(handler._action(attempts=2), 2)
        self.assertEqual(handler._action(attempts=3), 4)

    def test_crc32_check_propogates_error(self):
        handler = retryhandler.create_retry_handler(
            self.retry_config, operation_name='OperationFoo'
        )
        http_response = mock.Mock()
        http_response.status_code = 200
        # This is not the crc32 of b'foo', so this should
        # fail the crc32 check.
        http_response.headers = {'x-amz-crc32': 2356372768}
        http_response.content = b'foo'
        # The first 10 attempts we get a retry.
        self.assertEqual(
            handler(
                response=(http_response, {}),
                attempts=1,
                caught_exception=None,
                request_dict=REQUEST_DICT,
            ),
            1,
        )
        with self.assertRaises(ChecksumError):
            handler(
                response=(http_response, {}),
                attempts=10,
                caught_exception=None,
                request_dict=REQUEST_DICT,
            )


class TestRetryHandler(unittest.TestCase):
    def test_action_tied_to_policy(self):
        # When a retry rule matches we should return the
        # amount of time to sleep, otherwise we should return None.
        delay_function = retryhandler.create_exponential_delay_function(1, 2)
        checker = retryhandler.HTTPStatusCodeChecker(500)
        handler = retryhandler.RetryHandler(checker, delay_function)
        response = (HTTP_500_RESPONSE, {})

        self.assertEqual(
            handler(response=response, attempts=1, caught_exception=None), 1
        )
        self.assertEqual(
            handler(response=response, attempts=2, caught_exception=None), 2
        )
        self.assertEqual(
            handler(response=response, attempts=3, caught_exception=None), 4
        )
        self.assertEqual(
            handler(response=response, attempts=4, caught_exception=None), 8
        )

    def test_none_response_when_no_matches(self):
        delay_function = retryhandler.create_exponential_delay_function(1, 2)
        checker = retryhandler.HTTPStatusCodeChecker(500)
        handler = retryhandler.RetryHandler(checker, delay_function)
        response = (HTTP_200_RESPONSE, {})

        self.assertIsNone(
            handler(response=response, attempts=1, caught_exception=None)
        )


class TestCRC32Checker(unittest.TestCase):
    def setUp(self):
        self.checker = retryhandler.CRC32Checker('x-amz-crc32')

    def test_crc32_matches(self):
        http_response = mock.Mock()
        http_response.status_code = 200
        # This is the crc32 of b'foo', so this should
        # pass the crc32 check.
        http_response.headers = {'x-amz-crc32': 2356372769}
        http_response.content = b'foo'
        self.assertIsNone(
            self.checker(
                response=(http_response, {}),
                attempt_number=1,
                caught_exception=None,
            )
        )

    def test_crc32_missing(self):
        # It's not an error is the crc32 header is missing.
        http_response = mock.Mock()
        http_response.status_code = 200
        http_response.headers = {}
        self.assertIsNone(
            self.checker(
                response=(http_response, {}),
                attempt_number=1,
                caught_exception=None,
            )
        )

    def test_crc32_check_fails(self):
        http_response = mock.Mock()
        http_response.status_code = 200
        # This is not the crc32 of b'foo', so this should
        # fail the crc32 check.
        http_response.headers = {'x-amz-crc32': 2356372768}
        http_response.content = b'foo'
        with self.assertRaises(ChecksumError):
            self.checker(
                response=(http_response, {}),
                attempt_number=1,
                caught_exception=None,
            )


class TestDelayExponential(unittest.TestCase):
    def test_delay_with_numeric_base(self):
        self.assertEqual(
            retryhandler.delay_exponential(
                base=3, growth_factor=2, attempts=3
            ),
            12,
        )

    def test_delay_with_rand_string(self):
        delay = retryhandler.delay_exponential(
            base='rand', growth_factor=2, attempts=3
        )
        # 2 ** (3 - 1) == 4, so the retry is between 0, 4.
        self.assertTrue(0 <= delay <= 4)

    def test_value_error_raised_with_non_positive_number(self):
        with self.assertRaises(ValueError):
            retryhandler.delay_exponential(
                base=-1, growth_factor=2, attempts=3
            )


if __name__ == "__main__":
    unittest.main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tests import BaseSessionTest


class TestPolly(BaseSessionTest):
    def test_start_speech_synthesis_stream(self):
        """StartSpeechSynthesisStream operation removed due to h2 requirement"""
        polly_client = self.session.create_client('polly', 'us-west-2')
        try:
            polly_client.start_speech_synthesis_stream
        except AttributeError:
            pass
        else:
            self.fail(
                'start_speech_synthesis_stream shouldn\'t be available on the '
                'polly client.'
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

import pytest

from tests import requires_crt, skip_if_crt, skip_if_windows


class TestSkipIfWindows:
    def test_bare_skip_if_windows_fails_immediately(self):
        with pytest.raises(TypeError):

            @skip_if_windows
            def my_test():
                pass

    def test_skip_if_windows_skips_on_windows(self):
        with mock.patch('tests.platform') as mock_platform:
            mock_platform.system.return_value = 'Windows'

            @skip_if_windows('Not supported on Windows')
            def my_test():
                assert False

            assert getattr(my_test, '__unittest_skip__', False) is True

    def test_skip_if_windows_runs_on_non_windows(self):
        with mock.patch('tests.platform') as mock_platform:
            mock_platform.system.return_value = 'Linux'

            @skip_if_windows('Not supported on Windows')
            def my_test():
                pass

            assert getattr(my_test, '__unittest_skip__', False) is False


class TestRequiresCrt:
    def test_bare_requires_crt_fails_immediately(self):
        with pytest.raises(TypeError):

            @requires_crt
            def my_test():
                pass

    def test_requires_crt_skips_when_no_crt(self):
        with mock.patch('tests.HAS_CRT', False):

            @requires_crt()
            def my_test():
                assert False

            assert getattr(my_test, '__unittest_skip__', False) is True

    def test_requires_crt_runs_when_crt_available(self):
        with mock.patch('tests.HAS_CRT', True):

            @requires_crt()
            def my_test():
                pass

            assert getattr(my_test, '__unittest_skip__', False) is False


class TestSkipIfCrt:
    def test_bare_skip_if_crt_fails_immediately(self):
        with pytest.raises(TypeError):

            @skip_if_crt
            def my_test():
                pass

    def test_skip_if_crt_skips_when_crt_available(self):
        with mock.patch('tests.HAS_CRT', True):

            @skip_if_crt()
            def my_test():
                assert False

            assert getattr(my_test, '__unittest_skip__', False) is True

    def test_skip_if_crt_runs_when_no_crt(self):
        with mock.patch('tests.HAS_CRT', False):

            @skip_if_crt()
            def my_test():
                pass

            assert getattr(my_test, '__unittest_skip__', False) is False



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import botocore.session
from botocore.exceptions import ClientError
from tests import random_chars, unittest


class TestCloudformation(unittest.TestCase):
    def setUp(self):
        self.session = botocore.session.get_session()
        self.client = self.session.create_client('cloudformation', 'us-east-1')

    def test_handles_errors_with_template_body(self):
        # GetTemplate has a customization in handlers.py, so we're ensuring
        # it handles the case when a stack does not exist.
        with self.assertRaises(ClientError):
            self.client.get_template(
                StackName=f'does-not-exist-{random_chars(10)}'
            )


if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sysconfig

from tests import BaseClientDriverTest


class TestDoesNotLeakMemory(BaseClientDriverTest):
    # The user doesn't need to have credentials configured
    # in order to run the functional tests for resource leaks.
    # If we don't set this value and a user doesn't have creds
    # configured, each create_client() call will have to go through
    # the EC2 Instance Metadata provider's timeout, which can add
    # a substantial amount of time to the total test run time.
    INJECT_DUMMY_CREDS = True
    # We're making up numbers here, but let's say arbitrarily
    # that the memory can't increase by more than {SCALING_FACTOR} MBs.

    if sysconfig.get_config_var("Py_GIL_DISABLED") == 1:
        # Free-threaded Python objects require additional
        # memory for per-object locks and synchronization.
        SCALING_FACTOR = 17
    else:
        SCALING_FACTOR = 4
    MAX_GROWTH_BYTES = SCALING_FACTOR * 1024 * 1024

    def test_create_single_client_memory_constant(self):
        self.cmd('create_client', 's3')
        self.cmd('free_clients')
        self.record_memory()
        for _ in range(100):
            self.cmd('create_client', 's3')
            self.cmd('free_clients')
        self.record_memory()
        start, end = self.memory_samples
        self.assertTrue((end - start) < self.MAX_GROWTH_BYTES, (end - start))

    def test_create_memory_clients_in_loop(self):
        # We need to first create clients and free then before
        # recording our memory samples.  This is because of two reasons:
        # 1. Caching.  Some of the botocore internals will cache data, so
        #    the first client created will consume more memory than subsequent
        #    clients.  We're interested in growing memory, not total
        #    memory usage (for now), so we we care about the memory in the
        #    steady state case.
        # 2. Python memory allocation.  Due to how python allocates memory
        #    via it's small object allocator, arena's aren't freed until the
        #    entire 256kb isn't in use.  If a single allocation in a single
        #    pool in a single arena is still in use, the arena is not
        #    freed.  This case is easy to hit, and pretty much any
        #    fragmentation guarantees this case is hit.  The best we can
        #    do is verify that memory that's released back to python's
        #    allocator (but not to the OS) is at least reused in subsequent
        #    requests to create botocore clients.
        self.cmd('create_multiple_clients', '200', 's3')
        self.cmd('free_clients')
        self.record_memory()
        # 500 clients in batches of 50.
        for _ in range(10):
            self.cmd('create_multiple_clients', '50', 's3')
            self.cmd('free_clients')
        self.record_memory()
        start, end = self.memory_samples
        self.assertTrue((end - start) < self.MAX_GROWTH_BYTES, (end - start))

    def test_create_single_waiter_memory_constant(self):
        self.cmd('create_waiter', 's3', 'bucket_exists')
        self.cmd('free_waiters')
        self.record_memory()
        for _ in range(100):
            self.cmd('create_waiter', 's3', 'bucket_exists')
            self.cmd('free_waiters')
        self.record_memory()
        start, end = self.memory_samples
        self.assertTrue((end - start) < self.MAX_GROWTH_BYTES, (end - start))

    def test_create_memory_waiters_in_loop(self):
        # See ``test_create_memory_clients_in_loop`` to understand why
        # waiters are first initialized and then freed. Same reason applies.
        self.cmd('create_multiple_waiters', '200', 's3', 'bucket_exists')
        self.cmd('free_waiters')
        self.record_memory()
        # 500 waiters in batches of 50.
        for _ in range(10):
            self.cmd('create_multiple_waiters', '50', 's3', 'bucket_exists')
            self.cmd('free_waiters')
        self.record_memory()
        start, end = self.memory_samples
        self.assertTrue((end - start) < self.MAX_GROWTH_BYTES, (end - start))

    def test_create_single_paginator_memory_constant(self):
        self.cmd('create_paginator', 's3', 'list_objects')
        self.cmd('free_paginators')
        self.record_memory()
        for _ in range(100):
            self.cmd('create_paginator', 's3', 'list_objects')
            self.cmd('free_paginators')
        self.record_memory()
        start, end = self.memory_samples
        self.assertTrue((end - start) < self.MAX_GROWTH_BYTES, (end - start))

    def test_create_memory_paginators_in_loop(self):
        # See ``test_create_memory_clients_in_loop`` to understand why
        # paginators are first initialized and then freed. Same reason applies.
        self.cmd('create_multiple_paginators', '200', 's3', 'list_objects')
        self.cmd('free_paginators')
        self.record_memory()
        # 500 waiters in batches of 50.
        for _ in range(10):
            self.cmd('create_multiple_paginators', '50', 's3', 'list_objects')
            self.cmd('free_paginators')
        self.record_memory()
        start, end = self.memory_samples
        self.assertTrue((end - start) < self.MAX_GROWTH_BYTES, (end - start))



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
from unittest import mock

from botocore.plugin import (
    PluginContext,
    get_botocore_plugins,
    reset_plugin_context,
    set_plugin_context,
)


def test_get_botocore_plugins_ctx_takes_precedence():
    ctx = PluginContext(plugins="DISABLED")
    token = set_plugin_context(ctx)
    try:
        with mock.patch.dict(
            os.environ, {'BOTOCORE_EXPERIMENTAL__PLUGINS': 'a=b'}
        ):
            assert get_botocore_plugins() == 'DISABLED'
    finally:
        reset_plugin_context(token)



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from tests import ClientHTTPStubber, patch_load_service_model
from tests.functional.test_useragent import (
    get_captured_ua_strings,
    parse_registered_feature_ids,
)

MOCK_SERVICE_MODEL = {
    "version": "1.0",
    "documentation": "",
    "metadata": {
        "apiVersion": "2020-02-02",
        "endpointPrefix": "mockservice",
        "protocols": ["smithy-rpc-v2-cbor"],
        "protocol": "smithy-rpc-v2-cbor",
        "serviceFullName": "Mock Service",
        "serviceId": "mockservice",
        "signatureVersion": "v4",
        "signingName": "mockservice",
        "targetPrefix": "mockservice",
        "uid": "mockservice-2020-02-02",
    },
    "operations": {
        "MockOperation": {
            "name": "MockOperation",
            "http": {"method": "GET", "requestUri": "/"},
            "input": {"shape": "MockOperationRequest"},
            "documentation": "",
        },
    },
    "shapes": {
        "MockOpParam": {
            "type": "string",
        },
        "MockOperationRequest": {
            "type": "structure",
            "required": ["MockOpParam"],
            "members": {
                "MockOpParam": {
                    "shape": "MockOpParam",
                    "documentation": "",
                    "location": "uri",
                    "locationName": "param",
                },
            },
        },
    },
}

MOCK_RULESET = {
    "version": "1.0",
    "parameters": {},
    "rules": [
        {
            "conditions": [],
            "endpoint": {
                "url": "https://mockservice.us-west-2.amazonaws.com/"
            },
            "type": "endpoint",
        },
    ],
}


def test_user_agent_has_cbor_feature_id(patched_session, monkeypatch):
    patch_load_service_model(
        patched_session, monkeypatch, MOCK_SERVICE_MODEL, MOCK_RULESET
    )
    client = patched_session.create_client(
        'mockservice', region_name='us-west-2'
    )
    with ClientHTTPStubber(client) as stub_client:
        stub_client.add_response()
        # The mock CBOR operation registers `'PROTOCOL_RPC_V2_CBOR': 'M'`
        client.mock_operation(MockOpParam='mock-op-param-value')
    ua_string = get_captured_ua_strings(stub_client)[0]
    feature_list = parse_registered_feature_ids(ua_string)
    assert 'M' in feature_list
